@@ -1,6 +1,6 @@
 // fileread.cc -- read files for gold
 
-// Copyright 2006, 2007, 2008, 2009, 2010 Free Software Foundation, Inc.
+// Copyright 2006, 2007, 2008, 2009, 2010, 2011 Free Software Foundation, Inc.
 // Written by Ian Lance Taylor <iant@google.com>.
 
 // This file is part of gold.
@@ -24,10 +24,18 @@
 
 #include <cstring>
 #include <cerrno>
+#include <climits>
 #include <fcntl.h>
 #include <unistd.h>
+
+#ifdef HAVE_SYS_MMAN_H
 #include <sys/mman.h>
+#endif
+
+#ifdef HAVE_READV
 #include <sys/uio.h>
+#endif
+
 #include <sys/stat.h>
 #include "filenames.h"
 
@@ -41,8 +49,42 @@
 #include "gold-threads.h"
 #include "fileread.h"
 
+// For systems without mmap support.
+#ifndef HAVE_MMAP
+# define mmap gold_mmap
+# define munmap gold_munmap
+# ifndef MAP_FAILED
+#  define MAP_FAILED (reinterpret_cast<void*>(-1))
+# endif
+# ifndef PROT_READ
+#  define PROT_READ 0
+# endif
+# ifndef MAP_PRIVATE
+#  define MAP_PRIVATE 0
+# endif
+
+# ifndef ENOSYS
+#  define ENOSYS EINVAL
+# endif
+
+static void *
+gold_mmap(void *, size_t, int, int, int, off_t)
+{
+  errno = ENOSYS;
+  return MAP_FAILED;
+}
+
+static int
+gold_munmap(void *, size_t)
+{
+  errno = ENOSYS;
+  return -1;
+}
+
+#endif
+
 #ifndef HAVE_READV
-struct iovec { void* iov_base; size_t iov_len };
+struct iovec { void* iov_base; size_t iov_len; };
 ssize_t
 readv(int, const iovec*, int)
 {
@@ -53,6 +95,36 @@ readv(int, const iovec*, int)
 namespace gold
 {
 
+// Get the last modified time of an unopened file.
+
+bool
+get_mtime(const char* filename, Timespec* mtime)
+{
+  struct stat file_stat;
+
+  if (stat(filename, &file_stat) < 0)
+    return false;
+#ifdef HAVE_STAT_ST_MTIM
+  mtime->seconds = file_stat.st_mtim.tv_sec;
+  mtime->nanoseconds = file_stat.st_mtim.tv_nsec;
+#else
+  mtime->seconds = file_stat.st_mtime;
+  mtime->nanoseconds = 0;
+#endif
+  return true;
+}
+
+// Class File_read.
+
+// A lock for the File_read static variables.
+static Lock* file_counts_lock = NULL;
+static Initialize_lock file_counts_initialize_lock(&file_counts_lock);
+
+// The File_read static variables.
+unsigned long long File_read::total_mapped_bytes;
+unsigned long long File_read::current_mapped_bytes;
+unsigned long long File_read::maximum_mapped_bytes;
+
 // Class File_read::View.
 
 File_read::View::~View()
@@ -61,12 +133,17 @@ File_read::View::~View()
   switch (this->data_ownership_)
     {
     case DATA_ALLOCATED_ARRAY:
-      delete[] this->data_;
+      free(const_cast<unsigned char*>(this->data_));
       break;
     case DATA_MMAPPED:
       if (::munmap(const_cast<unsigned char*>(this->data_), this->size_) != 0)
         gold_warning(_("munmap failed: %s"), strerror(errno));
-      File_read::current_mapped_bytes -= this->size_;
+      if (!parameters->options_valid() || parameters->options().stats())
+	{
+	  file_counts_initialize_lock.initialize();
+	  Hold_optional_lock hl(file_counts_lock);
+	  File_read::current_mapped_bytes -= this->size_;
+	}
       break;
     case DATA_NOT_OWNED:
       break;
@@ -96,15 +173,6 @@ File_read::View::is_locked()
 
 // Class File_read.
 
-// A lock for the File_read static variables.
-static Lock* file_counts_lock = NULL;
-static Initialize_lock file_counts_initialize_lock(&file_counts_lock);
-
-// The File_read static variables.
-unsigned long long File_read::total_mapped_bytes;
-unsigned long long File_read::current_mapped_bytes;
-unsigned long long File_read::maximum_mapped_bytes;
-
 File_read::~File_read()
 {
   gold_assert(this->token_.is_writable());
@@ -409,34 +477,40 @@ File_read::make_view(off_t start, section_size_type size,
       gold_assert(psize >= size);
     }
 
-  File_read::View* v;
+  void* p;
+  View::Data_ownership ownership;
   if (byteshift != 0)
     {
-      unsigned char* p = new unsigned char[psize + byteshift];
+      p = malloc(psize + byteshift);
+      if (p == NULL)
+	gold_nomem();
       memset(p, 0, byteshift);
-      this->do_read(poff, psize, p + byteshift);
-      v = new File_read::View(poff, psize, p, byteshift, cache,
-                              View::DATA_ALLOCATED_ARRAY);
+      this->do_read(poff, psize, static_cast<unsigned char*>(p) + byteshift);
+      ownership = View::DATA_ALLOCATED_ARRAY;
     }
   else
     {
       this->reopen_descriptor();
-      void* p = ::mmap(NULL, psize, PROT_READ, MAP_PRIVATE,
-                       this->descriptor_, poff);
-      if (p == MAP_FAILED)
-	gold_fatal(_("%s: mmap offset %lld size %lld failed: %s"),
-		   this->filename().c_str(),
-		   static_cast<long long>(poff),
-		   static_cast<long long>(psize),
-		   strerror(errno));
-
-      this->mapped_bytes_ += psize;
-
-      const unsigned char* pbytes = static_cast<const unsigned char*>(p);
-      v = new File_read::View(poff, psize, pbytes, 0, cache,
-                              View::DATA_MMAPPED);
+      p = ::mmap(NULL, psize, PROT_READ, MAP_PRIVATE, this->descriptor_, poff);
+      if (p != MAP_FAILED)
+	{
+	  ownership = View::DATA_MMAPPED;
+	  this->mapped_bytes_ += psize;
+	}
+      else
+	{
+	  p = malloc(psize);
+	  if (p == NULL)
+	    gold_nomem();
+	  this->do_read(poff, psize, p);
+	  ownership = View::DATA_ALLOCATED_ARRAY;
+	}
     }
 
+  const unsigned char* pbytes = static_cast<const unsigned char*>(p);
+  File_read::View* v = new File_read::View(poff, psize, pbytes, byteshift,
+					   cache, ownership);
+
   this->add_view(v);
 
   return v;
@@ -494,7 +568,10 @@ File_read::find_or_make_view(off_t offset, off_t start,
     {
       gold_assert(aligned);
 
-      unsigned char* pbytes = new unsigned char[v->size() + byteshift];
+      unsigned char* pbytes;
+      pbytes = static_cast<unsigned char*>(malloc(v->size() + byteshift));
+      if (pbytes == NULL)
+	gold_nomem();
       memset(pbytes, 0, byteshift);
       memcpy(pbytes + byteshift, v->data() + v->byteshift(), v->size());
 
@@ -594,11 +671,22 @@ File_read::do_readv(off_t base, const Read_multiple& rm, size_t start,
 	       got, want, static_cast<long long>(base + first_offset));
 }
 
+// Portable IOV_MAX.
+
+#if !defined(HAVE_READV)
+#define GOLD_IOV_MAX 1
+#elif defined(IOV_MAX)
+#define GOLD_IOV_MAX IOV_MAX
+#else
+#define GOLD_IOV_MAX (File_read::max_readv_entries * 2)
+#endif
+
 // Read several pieces of data from the file.
 
 void
 File_read::read_multiple(off_t base, const Read_multiple& rm)
 {
+  static size_t iov_max = GOLD_IOV_MAX;
   size_t count = rm.size();
   size_t i = 0;
   while (i < count)
@@ -611,7 +699,7 @@ File_read::read_multiple(off_t base, const Read_multiple& rm)
       size_t j;
       for (j = i + 1; j < count; ++j)
 	{
-	  if (j - i >= File_read::max_readv_entries)
+	  if (j - i >= File_read::max_readv_entries || j - i >= iov_max / 2)
 	    break;
 	  const Read_multiple_entry& j_entry(rm[j]);
 	  off_t j_off = j_entry.file_offset;
@@ -694,7 +782,9 @@ File_read::clear_views(Clear_views_mode mode)
 	should_delete = false;
       else if (mode == CLEAR_VIEWS_ALL)
 	should_delete = true;
-      else if (p->second->should_cache() && keep_files_mapped)
+      else if ((p->second->should_cache()
+		|| p->second == this->whole_file_view_)
+	       && keep_files_mapped)
 	should_delete = false;
       else if (this->object_count_ > 1
       	       && p->second->accessed()
@@ -826,7 +916,7 @@ File_read::get_mtime()
 {
   struct stat file_stat;
   this->reopen_descriptor();
-  
+
   if (fstat(this->descriptor_, &file_stat) < 0)
     gold_fatal(_("%s: stat failed: %s"), this->name_.c_str(),
 	       strerror(errno));
@@ -837,8 +927,32 @@ File_read::get_mtime()
 #endif
 }
 
-// Open the file.
+// Try to find a file in the extra search dirs.  Returns true on success.
 
+bool
+Input_file::try_extra_search_path(int* pindex,
+				  const Input_file_argument* input_argument,
+				  std::string filename, std::string* found_name,
+				  std::string* namep)
+{
+  if (input_argument->extra_search_path() == NULL)
+    return false;
+
+  std::string name = input_argument->extra_search_path();
+  if (!IS_DIR_SEPARATOR(name[name.length() - 1]))
+    name += '/';
+  name += filename;
+
+  struct stat dummy_stat;
+  if (*pindex > 0 || ::stat(name.c_str(), &dummy_stat) < 0)
+    return false;
+
+  *found_name = filename;
+  *namep = name;
+  return true;
+}
+
+// Find the actual file.
 // If the filename is not absolute, we assume it is in the current
 // directory *except* when:
 //    A) input_argument_->is_lib() is true;
@@ -848,34 +962,37 @@ File_read::get_mtime()
 // the file location, rather than the current directory.
 
 bool
-Input_file::open(const Dirsearch& dirpath, const Task* task, int *pindex)
+Input_file::find_file(const Dirsearch& dirpath, int* pindex,
+		      const Input_file_argument* input_argument,
+		      bool* is_in_sysroot,
+		      std::string* found_name, std::string* namep)
 {
   std::string name;
 
   // Case 1: name is an absolute file, just try to open it
   // Case 2: name is relative but is_lib is false, is_searched_file is false,
   //         and extra_search_path is empty
-  if (IS_ABSOLUTE_PATH(this->input_argument_->name())
-      || (!this->input_argument_->is_lib()
-	  && !this->input_argument_->is_searched_file()
-	  && this->input_argument_->extra_search_path() == NULL))
+  if (IS_ABSOLUTE_PATH(input_argument->name())
+      || (!input_argument->is_lib()
+	  && !input_argument->is_searched_file()
+	  && input_argument->extra_search_path() == NULL))
     {
-      name = this->input_argument_->name();
-      this->found_name_ = name;
+      name = input_argument->name();
+      *found_name = name;
+      *namep = name;
+      return true;
     }
   // Case 3: is_lib is true or is_searched_file is true
-  else if (this->input_argument_->is_lib()
-	   || this->input_argument_->is_searched_file())
+  else if (input_argument->is_lib()
+	   || input_argument->is_searched_file())
     {
-      // We don't yet support extra_search_path with -l.
-      gold_assert(this->input_argument_->extra_search_path() == NULL);
       std::string n1, n2;
-      if (this->input_argument_->is_lib())
+      if (input_argument->is_lib())
 	{
 	  n1 = "lib";
-	  n1 += this->input_argument_->name();
+	  n1 += input_argument->name();
 	  if (parameters->options().is_static()
-	      || !this->input_argument_->options().Bdynamic())
+	      || !input_argument->options().Bdynamic())
 	    n1 += ".a";
 	  else
 	    {
@@ -884,49 +1001,70 @@ Input_file::open(const Dirsearch& dirpath, const Task* task, int *pindex)
 	    }
 	}
       else
-	n1 = this->input_argument_->name();
-      name = dirpath.find(n1, n2, &this->is_in_sysroot_, pindex);
+	n1 = input_argument->name();
+
+      if (Input_file::try_extra_search_path(pindex, input_argument, n1,
+					    found_name, namep))
+        return true;
+
+      if (!n2.empty() && Input_file::try_extra_search_path(pindex,
+							   input_argument, n2,
+							   found_name, namep))
+        return true;
+
+      // It is not in the extra_search_path.
+      name = dirpath.find(n1, n2, is_in_sysroot, pindex);
       if (name.empty())
 	{
 	  gold_error(_("cannot find %s%s"),
-	             this->input_argument_->is_lib() ? "-l" : "",
-		     this->input_argument_->name());
+	             input_argument->is_lib() ? "-l" : "",
+		     input_argument->name());
 	  return false;
 	}
       if (n2.empty() || name[name.length() - 1] == 'o')
-	this->found_name_ = n1;
+	*found_name = n1;
       else
-	this->found_name_ = n2;
+	*found_name = n2;
+      *namep = name;
+      return true;
     }
   // Case 4: extra_search_path is not empty
   else
     {
-      gold_assert(this->input_argument_->extra_search_path() != NULL);
+      gold_assert(input_argument->extra_search_path() != NULL);
 
-      // First, check extra_search_path.
-      name = this->input_argument_->extra_search_path();
-      if (!IS_DIR_SEPARATOR (name[name.length() - 1]))
-        name += '/';
-      name += this->input_argument_->name();
-      struct stat dummy_stat;
-      if (*pindex > 0 || ::stat(name.c_str(), &dummy_stat) < 0)
+      if (try_extra_search_path(pindex, input_argument, input_argument->name(),
+                                found_name, namep))
+        return true;
+
+      // extra_search_path failed, so check the normal search-path.
+      int index = *pindex;
+      if (index > 0)
+        --index;
+      name = dirpath.find(input_argument->name(), "",
+                          is_in_sysroot, &index);
+      if (name.empty())
         {
-          // extra_search_path failed, so check the normal search-path.
-	  int index = *pindex;
-	  if (index > 0)
-	    --index;
-          name = dirpath.find(this->input_argument_->name(), "",
-			      &this->is_in_sysroot_, &index);
-          if (name.empty())
-            {
-              gold_error(_("cannot find %s"),
-			 this->input_argument_->name());
-	      return false;
-            }
-	  *pindex = index + 1;
+          gold_error(_("cannot find %s"),
+                     input_argument->name());
+          return false;
         }
-      this->found_name_ = this->input_argument_->name();
+      *found_name = input_argument->name();
+      *namep = name;
+      *pindex = index + 1;
+      return true;
     }
+}
+
+// Open the file.
+
+bool
+Input_file::open(const Dirsearch& dirpath, const Task* task, int* pindex)
+{
+  std::string name;
+  if (!Input_file::find_file(dirpath, pindex, this->input_argument_,
+			     &this->is_in_sysroot_, &this->found_name_, &name))
+    return false;
 
   // Now that we've figured out where the file lives, try to open it.
 
