@@ -725,7 +725,6 @@ parse_args (unsigned argc, char **argv)
     {
       int longind;
       int optc;
-      lang_input_statement_type *input;
 
       /* Using last_optind lets us avoid calling ldemul_parse_args
 	 multiple times on a single option, which would lead to
@@ -764,10 +763,7 @@ parse_args (unsigned argc, char **argv)
 	  einfo (_("%P%F: use the --help option for usage information\n"));
 
 	case 1:			/* File name.  */
-	  input = lang_add_input_file (optarg,
-				       lang_input_file_is_file_enum, NULL);
-	  cmdline_option_append (cmdline_is_lang_input_statement_enum,
-				 input);
+	  lang_add_input_file (optarg, lang_input_file_is_file_enum, NULL);
 	  break;
 
 	case OPTION_IGNORE:
@@ -924,10 +920,7 @@ parse_args (unsigned argc, char **argv)
 	  ldfile_add_library_path (optarg, TRUE);
 	  break;
 	case 'l':
-	  input = lang_add_input_file (optarg,
-				       lang_input_file_is_l_enum, NULL);
-	  cmdline_option_append (cmdline_is_lang_input_statement_enum,
-				 input);
+	  lang_add_input_file (optarg, lang_input_file_is_l_enum, NULL);
 	  break;
 	case 'M':
 	  config.map_filename = "-";
@@ -1112,11 +1105,9 @@ parse_args (unsigned argc, char **argv)
 	    if (stat (optarg, &s) >= 0
 		&& ! S_ISDIR (s.st_mode))
 	      {
-		input = lang_add_input_file (optarg,
-					     lang_input_file_is_symbols_only_enum,
-					     NULL);
-		cmdline_option_append (cmdline_is_lang_input_statement_enum,
-				       input);
+		lang_add_input_file (optarg,
+				     lang_input_file_is_symbols_only_enum,
+				     NULL);
 		break;
 	      }
 	  }
@@ -1248,13 +1239,11 @@ parse_args (unsigned argc, char **argv)
 	  trace_files = TRUE;
 	  break;
 	case 'T':
-	  cmdline_linker_script = 1;
 	  previous_script_handle = saved_script_handle;
 	  ldfile_open_command_file (optarg);
 	  parser_input = input_script;
 	  yyparse ();
 	  previous_script_handle = NULL;
-	  cmdline_linker_script = 0;
 	  break;
 	case OPTION_DEFAULT_SCRIPT:
 	  command_line.default_script = optarg;
@@ -1504,14 +1493,14 @@ parse_args (unsigned argc, char **argv)
 	  command_line.accept_unknown_input_arch = FALSE;
 	  break;
 	case '(':
-	  lang_enter_group (TRUE);
+	  lang_enter_group ();
 	  ingroup++;
 	  break;
 	case ')':
 	  if (! ingroup)
 	    einfo (_("%P%F: group ended before it began (--help for usage)\n"));
 
-	  lang_leave_group (TRUE);
+	  lang_leave_group ();
 	  ingroup--;
 	  break;
 
@@ -1545,7 +1534,7 @@ parse_args (unsigned argc, char **argv)
 
   while (ingroup)
     {
-      lang_leave_group (FALSE);
+      lang_leave_group ();
       ingroup--;
     }
 
